#define _BSD 1
#include <sys/types.h>
#include <ppa.h>
#define _SOCKADDR_LEN
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <sys/time.h>
#include <table.h>
#include <sys/fcntl.h>
#include <signal.h>
#include <math.h>
#include <stdio.h>

table probes;
table edges;
unsigned int total_in_flight=0;
unsigned int ndt=0;
unsigned int probe_count=1;
unsigned int cumulative=0;
unsigned int where=0;
unsigned int source_address;
unsigned int timeout=5;
unsigned int concurrent_probes=30;
unsigned int print_rx=0;
unsigned int print_tx=0;
unsigned int edge_logging=0;
unsigned int bins=15;
unsigned int reset_distribution_on_dump=1;
unsigned int mail_interval=3600;
char *mailto=0;

#define WEIGHT 0.000001 
#define CURVE 0.3
#define MIN(x,y) ((x)<(y)?(x):(y))
#define BIN_FUNCTION(x) ((int)MIN(bins-1,\
				  (pow(((float)x)*WEIGHT,CURVE)*bins)))
#define INVERSE_BIN_FUNCTION(x) ((int)(pow((float)x/(float)bins,1.0/CURVE)/WEIGHT))

typedef struct edge *edge;
typedef struct probe *probe;
typedef struct path *path;
struct edge {
  unsigned int from;
  unsigned int to;
  unsigned int last_seen;
  unsigned int *delay;
  unsigned int interval; 
  unsigned int last_sent_seconds;
  unsigned int last_sent_microseconds;
  unsigned int total_sent;
  unsigned int total_acknowledged;
  edge *last;
  edge next;
  unsigned int references; /* paths containing this edge*/
  unsigned int average;
  ppa p;
  double sum;
  double sum_squares;
};

edge edge_ring=0;
edge edge_pointer=0;

struct path {
  edge e;
  path next;
  path *last;
};

struct probe {
  unsigned int last_sent_seconds;
  unsigned int last_sent_microseconds;
  unsigned int number_retries;
  unsigned int probe_number;
  unsigned int ttl;
  unsigned int probe_sequence;
  unsigned int last_sequence;
  unsigned int dest;
  unsigned int last_dest;
  unsigned int retry_limit;
  unsigned int sweep_lower_limit;
  unsigned int sweep_upper_limit;
  unsigned int sweep_step;
  unsigned int probe_failures;
  path p;
  path *current;
};

struct opacket {
    struct ip ip;
    unsigned int probe_number;
    unsigned int probe_sequence;
  /*    unsigned seconds;
    unsigned microseconds;*/
};

struct ping_packet {
    struct ip ip;

    unsigned char type;
    unsigned char code;
    unsigned short checksum;
    unsigned short identifier;
    unsigned short sequence_number;

    unsigned int edge_src;
    unsigned int edge_dst;
    unsigned int seconds;
    unsigned int microseconds;
};


static int probe_compare(probe p,unsigned int key)
{
    return(p->probe_number==key);
}

static int key_from_edge(unsigned int *pair)
{
  return(key_from_int(pair[0]^(pair[1]<<1)));
}

static int edge_compare(edge e,unsigned int *pair)
{
    return((pair[0]==e->from) && (pair[1]==e->to));
}

edge create_edge(unsigned int from, unsigned int to)
{
  unsigned int f,t,i;
  unsigned int et[2];

  edge e=(edge)malloc(sizeof(struct edge));
  e->total_sent=0;
  et[0]=e->from=from;
  et[1]=e->to=to;
  e->p=create_ppa(10);
  if (ndt) {
    t=htonl(e->to);
    f=htonl(e->from);
    printf ("(\"srcip=%d.%d.%d.%d\", \"dstip=%d.%d.%d.%d\")\n",
	    (f>>24),(f>>16)&255,(f>>8)&255,f&255,
	    (t>>24),(t>>16)&255,(t>>8)&255,t&255);
    fflush(stdout);
  }
  /*  e->delay=(unsigned int *)malloc(sizeof(unsigned int)*bins);*/
  e->interval=10;
  e->sum=e->sum_squares=0.0;
  e->total_acknowledged=0;
  /*  for (i=0;i<bins;i++) e->delay[i]=0;*/
  table_insert(edges,e,et);
  if (edge_ring) {
    e->last=edge_ring->last;
    e->next=edge_ring;
    *edge_ring->last=e;
    edge_ring->last=&e->next;
  } else {
    e->last=&e->next;
    e->next=e;
    edge_pointer=e;
  }
  edge_ring=e;
  return(e);
}

void reset_probe(probe p)
{
  p->ttl=1;
  p->number_retries=0;
  p->probe_sequence=0;
  p->retry_limit=10;
  p->last_dest=source_address;
  p->current=&p->p;
}

probe create_probe (unsigned int destination)
{
  probe p=(probe)malloc(sizeof(struct probe));
  p->probe_number=probe_count++;
  p->dest=destination;
  p->p=0;
  table_insert(probes,p,p->probe_number);
  reset_probe(p);
  return(p);
}


static char *hex_digits="0123456789abcdef";

void print_buffer(char *p,void *v ,int length)
{
  unsigned char *n=v;
  int count,len,i;
  unsigned int trash;
  unsigned char z,y;
  char s[3];

  s[2]=0;
  if (n && p) {
    for (count=0;count<length;count++){
      if (!(count%24)) printf ("\n%s",p);
      s[0]=hex_digits[(n[count]>>4)&15];
      s[1]=hex_digits[n[count]&15];
      printf ("%s",s);
      if (!((count+1)%4)) printf (" ");
    }
    printf ("\n");
  }
}

int send_probe(probe p,int k[3])
{
  struct opacket op;
  struct sockaddr to;
  struct ip *ip = &op.ip;
  int i;
  unsigned int seconds=k[0];
  unsigned int microseconds=k[1];
  int fd=k[2];
  unsigned char c;
  
  if ((p->last_sent_seconds+timeout < seconds) || (p->number_retries==0)){
    if (p->number_retries++<p->retry_limit){
      
      ip->ip_off = 0;
      ip->ip_hl = sizeof(*ip) >> 2;
      ip->ip_p = 254; /* hopefully unknown protocol */
      
      ip->ip_len = sizeof(struct opacket);
      ip->ip_ttl = p->ttl;
      ip->ip_v = 4;
      ip->ip_id = /*htons(f->ident+f->count);*/0;
      ip->ip_src.s_addr =source_address;
      ip->ip_dst.s_addr =p->dest;
      op.probe_number=htonl(p->probe_number);
      op.probe_sequence=htonl(p->probe_sequence);
      /*	op.seconds=seconds;
		op.microseconds=microseconds;*/
      
      p->last_sent_seconds=seconds;
      p->last_sent_microseconds=microseconds;
      total_in_flight++;
      if (print_tx) print_buffer("v   ",&op,sizeof(op));
      if (sendto(fd, &op, sizeof(op), 0, &to, sizeof(struct sockaddr))!=
	  sizeof(op)) perror ("Send error");
    } else reset_probe(p); /* this should continue on, as
			      traceroute would, but being
			      conservative*/
  }
  return(1);
}

int edge_probe (edge e,int k[3])
{
  struct ping_packet ping;
  struct sockaddr to;
  struct ip *ip = &ping.ip;
  int i;
  unsigned int seconds=k[0];
  unsigned int n,t,f;
  unsigned int microseconds=k[1];
  int fd=k[2];
  unsigned char c;
  
  ip->ip_off = 0;
  ip->ip_hl = sizeof(*ip) >> 2;
  ip->ip_p = 1;
  
  ip->ip_len = sizeof(struct ping_packet);
  ip->ip_ttl = 0xfe;
  ip->ip_v = 4;
  ip->ip_id = 0;
  ip->ip_src.s_addr =source_address;
  ip->ip_dst.s_addr =e->to;
  
  ping.type=8;
  ping.code=0;
  ping.identifier=ping.sequence_number=0;
  ping.seconds=htonl(seconds);
  ping.microseconds=htonl(microseconds);
  ping.edge_src=e->from;
  ping.edge_dst=e->to;
  
  n=0x800;
  n+=(seconds>>16)+(seconds&65535);
  n+=(microseconds>>16)+(microseconds&65535);
  t=htonl(e->to);
  f=htonl(e->from);
  n+=(t>>16)+(t&65535);
  n+=(f>>16)+(f&65535);
  
  while (n>65535) n=(n>>16)+(n&65535);
  ping.checksum=~n;
  ping.checksum=htons(ping.checksum);
  e->last_sent_seconds=seconds;
  e->last_sent_microseconds=microseconds;
  e->total_sent++;
  if (print_tx) print_buffer("v   ",&ping,sizeof(struct ping_packet));
  if (sendto(fd, &ping, sizeof(ping), 0, &to,sizeof(struct sockaddr))!=
      sizeof(struct ping_packet)) perror ("Send error");
}
    
unsigned int last_dump;
typedef struct dump {
  void (*write)();
  void *context;
} *dump;

void each_dump(edge e,dump d)
{
  char oline[512]; 
  unsigned int i;
  unsigned int from=htonl(e->from);
  unsigned int to=htonl(e->to);
 
  sprintf (oline,
	   "(\"srcip=%d.%d.%d.%d\", \"dstip=%d.%d.%d.%d\"",
	   (from>>24),
	   (from>>16)&255,
	   (from>>8)&255,
	   from&255,
	   
	   (to>>24),
	   (to>>16)&255,
	   (to>>8)&255,
	   to&255);
   if (e->total_sent)
    sprintf (oline,"%s,\"acked=%f\"",oline,
	   ((float)e->total_acknowledged/(float)e->total_sent));
  if (e->total_acknowledged > 11 ){
    sprintf (oline,"%s,\"delays=",oline);
    for (i=1;i<11;i++) sprintf (oline,"%s%f:",oline,e->p->markers[i]/1000.0);
    sprintf (oline,"%s%f\"",oline,e->p->markers[11]/1000.0);
  }
  if (e->total_acknowledged>1){
    sprintf (oline,"%s,\"variation=%lg\"",oline,
	     sqrt(((double)e->total_acknowledged*e->sum_squares-
		   e->sum*e->sum)/((double)e->total_acknowledged*
				   (double)(e->total_acknowledged-1))));

  }
  sprintf (oline,"%s,\"samples=%d\")\n",oline,e->total_sent);
  (*d->write)(d->context,oline,strlen(oline));
  if (!cumulative){
    destroy_ppa(e->p);
    e->p=create_ppa(10);
    e->total_sent=0;
    e->sum=0;
    e->sum_squares=0;
    e->total_acknowledged=0;
  }
}

unsigned int address_formatter(char *n)
{
    unsigned int a,b,c,d;
    sscanf(n,"%d.%d.%d.%d\n",&a,&b,&c,&d);
    return ((a<<24)|(b<<16)|(c<<8)|d);
}

void dump_state(void (*write)(),void *context)
{
  struct dump d;
  d.write=write;
  d.context=context;
  iterate_table_entries(edges,each_dump,&d);	
}

extern void write();

void dump_edges(char *filename)
{
  int fd;
  if ((fd=open(filename,O_WRONLY|O_CREAT|O_TRUNC,0664)) > 0){
    dump_state(write,fd);
    close (fd);
  } else printf ("couldn't open %s\n",filename);
}

unsigned int parse_ip(char *v)
{
  unsigned int a,b,c,d;
  sscanf(v,"%d.%d.%d.%d",&a,&b,&c,&d);
  a=((a<<24)|(b<<16)|(c<<8)|d);
  return(htonl(a));
}

void sigint()
{
  int i;
  int k;
  printf ("#edge list:\n#");
  /*  for (i=0;i<bins;i++){
    k=INVERSE_BIN_FUNCTION(i);
    printf ("%d.%d ",k/1000,(k%1000));
  }
  printf ("\n");*/
  dump_state(write,0);
  exit(1);
}

int receive_probe(unsigned int *rx,int length,unsigned int seconds,
		   unsigned int microseconds)
{
  probe p;
  path k,j;
  edge e;
  unsigned int et[2];

  if (p=(probe)table_find(probes,htonl(rx[12]))){
    et[0]=p->last_dest;
    et[1]=rx[3];
    if (p->last_dest == rx[3]){
      if (p->probe_sequence == htonl(rx[13])){
	if (p->sweep_step && ((p->dest+p->sweep_step) < p->sweep_upper_limit)){
	  p->ttl=1;
	  p->dest+=p->sweep_step;
	  p->number_retries=0;
	  p->probe_sequence=0;
	} else {
	  /* should really shedule for periodic update*/
	  /* just reset and crank up time*/
	  /* printf ("removing %x\n",htonl(rx[12]));*/
	  table_remove(probes,htonl(rx[12]));
	}
      }
    } else {
      if (!(e=(edge)table_find(edges,&et))) 
	e=create_edge(p->last_dest,rx[3]);

      if (*p->current){
	if ((*p->current)->next)
	  if ((*p->current)->next->e!=e){
	    /* this currently kills all the upstream
	       children*/
	    printf ("#routing change\n");
	    for (k=(*p->current)->next;k;k=j){
	      j=k->next;
	      k->e->references--;
	      free(k);
	    }
	    (*p->current)->e=e;
	    e->references++;
	    (*p->current)->next=0;
	  }
      } else { 
	k=(path)malloc(sizeof(struct path));
	k->last=p->current;
	e->references++;
	if (k->next=*p->current) (*p->current)->last=&k->next;
	*p->current=k;
      }
      e->last_seen=seconds;
      p->ttl++;
      p->probe_sequence++;
      p->last_dest=rx[3];
      p->number_retries=0;
    }
    return(1);
  }
  return(0);
}

void receive_edge_probe(unsigned int *rx,int length,
			unsigned int seconds,
			unsigned int microseconds)
{
  edge e;
  unsigned int et[2];
  unsigned int bin;
  unsigned int delta;
  unsigned int from,to;
  from=htonl(et[0]=rx[7]);
  to=htonl(et[1]=rx[8]);
  if (e=(edge)table_find(edges,et)){
    e->total_acknowledged++;
    delta=(int)(seconds-htonl(rx[9]))*1000000+
      ((int)microseconds-(int)htonl(rx[10]));
    /*    e->delay[BIN_FUNCTION(delta)]++;  */
    add_sample(e->p,delta);
    delta/=1000.0;
    e->sum+=(double)delta;
    e->sum_squares+=(double)delta*(double)delta;
    if (edge_logging){
      printf ("%d %d %d %d: %d %d\n",seconds,htonl(rx[9]), 
	      microseconds, htonl(rx[10]), 
	      delta,BIN_FUNCTION(delta));
      
      printf ("%d.%d.%d.%d %d.%d.%d.%d: %d.%d\n",
	      (from>>24),
	      (from>>16)&255,
	      (from>>8)&255,
	      from&255,
	      (to>>24),
	      (to>>16)&255,
	      (to>>8)&255,
	      to&255,delta/1000,delta%1000);
    }
  }
}

void int_set(int argc,int *context,char *n,char *real)
{
  *context=atoi(real);
}

void commit_probe_string(int argc,void *foo,void *command,char *name)
{
  if (argc) create_probe(parse_ip(name));
}

void commit_sweep_string(int argc,void *foo,char *start,
			 char *end, char *interval)
{
  probe p;
  if (argc) p=create_probe(parse_ip(start));
}

void commit_source_address(int argc,unsigned int *address,void *foo,
			   char *name)
{ 
  *address=parse_ip(name);
}

void commit_mailto(int argc,char **address,char *name,char *real)
{ 
  *address=strdup(real);
}
extern void write_mail();
void advance_probes(unsigned int *state)
{
  static int probe_last=0;
  void *n;
  static unsigned int lastmail_seconds=0;
  if ((edge_ring==edge_pointer) && (!probe_last)){
    probe_last=1;
    total_in_flight=0;
    iterate_table_entries(probes,send_probe,state);
    if (!lastmail_seconds) lastmail_seconds=state[0];
    if (mailto && ((mail_interval+lastmail_seconds)<state[0])){
      lastmail_seconds=state[0];
      n=open_mail("sendmail",mailto);
      write_mail(n,"To: ",4);
      write_mail(n,mailto,strlen(mailto));
      write_mail(n,"\nSubject: squeeb\n",17);
      dump_state(write_mail,n);
      close_mail(n);
    }
  } else {
    probe_last=0;
    if (edge_pointer){ 
      edge_probe(edge_pointer,state);
      edge_pointer=edge_pointer->next;
    }
  }
}

main(int argc, char **argv)
{
  int s,fromlen;
  unsigned int state[3],s_send;
  unsigned int rx[1024];
  unsigned int i,haddr; 
  struct timeval t,wait;
  struct sockaddr from;
  void *n;
  fd_set fds;
  int cc = 0,on=1;
  void (*commit)()=0;
  void *commit_context;
  int commit_count,commit_place;

  probes=create_table(probe_compare,key_from_int);
  edges=create_table(edge_compare,key_from_edge);  

  for (i=1;i<argc;i++){ 
    if (argv[i][0]=='-'){
      if (commit){
	(*commit)(commit_count,commit_context,
		  argv[commit_place],argv[commit_place+1],argv[commit_place+2]); 
	commit=0;
      }
      if (strcmp(argv[i]+1,"sweep")==0) commit=commit_sweep_string;
      if (strcmp(argv[i]+1,"timeout")==0){
	commit=int_set;
	commit_context=&timeout;
      }
      if (strcmp(argv[i]+1,"bins")==0) {
	commit=int_set;
	commit_context=&bins;
      }
      if (strcmp(argv[i]+1,"src")==0) {
	commit=commit_source_address;
	commit_context=&source_address;
      }
      if (strcmp(argv[i]+1,"ndt")==0) ndt=1;
      if (strcmp(argv[i]+1,"edge_log")==0) edge_logging=1;
      if (strcmp(argv[i]+1,"ptx")==0) print_tx=1;
      if (strcmp(argv[i]+1,"prx")==0) print_rx=1;
      if (strcmp(argv[i]+1,"mail")==0) {
	commit_context=&mailto;
	commit=commit_mailto;
      }
      if (strcmp(argv[i]+1,"mailinterval")==0) {
	commit_context=&mail_interval;
	commit=int_set;
      }
      if (strcmp(argv[i]+1,"dest")==0) commit=commit_probe_string;
      commit_place=i;
    } else commit_count++;
  }
  if (commit)
    (*commit)(commit_count,commit_context,argv[commit_place],
	      argv[commit_place+1]); 

  signal(SIGINT,sigint);
  bzero (&from,sizeof(struct sockaddr));
  from.sa_family = AF_INET;
  from.sa_len=6;

  bcopy (&haddr, &((struct sockaddr_in *)&from)->sin_addr, 4);
  if ((s=socket(AF_INET,SOCK_RAW,IPPROTO_ICMP)) < 0){
      perror ("couldn't open raw icmp socket");
      exit(0);
  }
  if ((s_send=socket(AF_INET,SOCK_RAW,IPPROTO_RAW)) < 0){
    perror ("couldn't open raw socket");
    exit(0);
  }
  if (setsockopt(s_send, IPPROTO_IP, IP_HDRINCL, (char *)&on,
		 sizeof(on)) < 0) {
    perror("traceroute: IP_HDRINCL");
    exit(6);
  }
  if (bind(s_send, (struct sockaddr *)&from, sizeof(from)) < 0) {
    perror ("bind");
    exit (1);
  }

  gettimeofday(&t,0);
  state[0]=t.tv_sec;
  state[1]=t.tv_usec;
  state[2]=s_send;

  advance_probes(state);
  while (1){    

    FD_ZERO(&fds);
    FD_SET(s, &fds);
    wait.tv_sec = timeout; wait.tv_usec = 0;
    
    if ((cc=select(32, &fds, (fd_set *)0, (fd_set *)0, &wait)) > 0){
      if (FD_ISSET(s,&fds)){
	if ((cc=recvfrom(s,rx, sizeof(rx), 0, &from, &fromlen)) > 0){
	  gettimeofday(&t, 0);	
	  if (print_rx)  print_buffer("^ ",rx,cc);
	  
	  state[0]=t.tv_sec;
	  state[1]=t.tv_usec;
	  /* should also mux on code */
	  if ((cc==56) && (((htonl(rx[2])>>16)&255)==1)){
	    if (receive_probe(rx,cc,state[0],state[1])){
	      if (!--total_in_flight) advance_probes(state);
	    }
	  } else {
	    if ((cc==44) && (((htonl(rx[2])>>16)&255)==1)){
	      receive_edge_probe(rx,cc,state[0],state[1]);
	      advance_probes(state);
	    }
	  }
	}
      }
    } else {
      if (!cc){
	gettimeofday(&t,0);
	state[0]=t.tv_sec;
	state[1]=t.tv_usec;
	advance_probes(state);
      }
    }
  }
}
