static char rcsid[]="$Id: table.c,v 1.1 1996/05/02 04:09:44 yuri Exp $";
#include <table.h>

typedef struct timeout_entry {
  table t;
  int timer_id;
  void *comparator;
} *timeout_entry;


unsigned int key_from_int(int i)
{
  return(i);
}


table create_table (int (*compare_function)(),
		    unsigned int (*key_function)())
{
  table new=(table)malloc(sizeof(struct table));
  int i;
  new->compare_function=compare_function;
  new->key_function=key_function;
  new->number_of_entries=0;
  new->size=64;
  new->entries=(table_entry *)malloc(sizeof(table_entry)*new->size);
  for (i=0;i<64;i++) new->entries[i]=0; 
  new->key_multiplier=56377;
  return(new);
}

static table_entry *table_lookup (table t,void *comparator,
				  unsigned int k,
				  int (*compare_function)(),
				  int *success)
{
  unsigned int key=(k*t->key_multiplier)%t->size;
  table_entry *i;

  for (i=&(t->entries[key]);*i;i=&((*i)->next)){
    if (compare_function && ((*i)->key==k))
      if ((*t->compare_function)((*i)->value,comparator)){
	*success=1;
	return(i);
      }
  }
  *success=0;
  return(&(t->entries[key]));
}

void *table_find (table t, void *comparator)
{
  int success;
  table_entry temp;
  table_entry* entry=table_lookup(t,comparator,
				  (*t->key_function)(comparator),
				  t->compare_function,
				  &success);
  if (success)  return((*entry)->value);
  return(0);
}


static void resize_table(table t, int size)
{
  int old_size=t->size;
  table_entry *old_entries=t->entries;
  int i; 
  table_entry j,n;
  table_entry *position,entry;
  int success;
  
  t->size=size;
  /* generate a prime here t->key_multiplier = size+3;*/
  t->number_of_entries=0;
  t->entries=(table_entry *)
    malloc(sizeof(table_entry)*t->size);
  /*  if (!t->entries) eerror("failed to malloc new table.\n");*/
  for (i=0;i<t->size;i++) t->entries[i]=0; 
  for (i=0;i<old_size;i++)
    for (j=old_entries[i];j;j=n){
      n=j->next;
      position=table_lookup(t,0,j->key,0,&success);
      j->next= *position;
      *position=j;
    }
  free(old_entries);
}


void table_insert (table t, void *value, void *comparator)
{
  int success;
  unsigned int k=(*t->key_function)(comparator);
  table_entry *position=table_lookup(t,comparator,k,
				     t->compare_function,&success);
  table_entry entry;
  int new;

  if (success) {
    entry = *position;
  } else {
    entry = (table_entry)malloc(sizeof(struct table_entry));
    entry->next= *position;
    *position=entry;
    if (t->number_of_entries++ > 2*t->size) resize_table(t,t->size*2);
  }
  entry->expiration=0;
  entry->value=value;
  entry->key=k;
}

void table_remove (table t, void *comparator)
{
  int success;
  table_entry temp;
  table_entry *position=table_lookup(t,comparator,
				     (*t->key_function)(comparator),
				     t->compare_function,&success);
  if(success) {
    temp=*position;
    *position=(*position)->next;
    free(temp);
    t->number_of_entries--;
  }
}


/* this will break if operations on the table occur in the handler*/
/* should take a retain/remove handler result*/
void iterate_table_entries(table t, int (*handler)(), void *arg)
{
  int i;
  table_entry *j,*next;
  
  for (i=0;i<t->size;i++)
    for (j=t->entries+i;*j;j=next){
	next=&((*j)->next);
	if (!(*handler)((*j)->value,arg)){
	    *j=(*j)->next;
	    t->number_of_entries--;
	    free(*j);
	}
    }
}

void destroy_table(table t)
{
  int i;
  table_entry j,k;
  for (i=0;i<t->size;i++)
    for (j=t->entries[i];j;j=k){
      k=j->next;
      free(j);
    }
  free(t->entries);
  free(t);
}
